/*
   atarisio V0.06 beta
   a kernel module for handling the Atari 8bit SIO protocol

   Copyright (C) 2002-2005 Matthias Reichl <hias@horus.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/config.h>
#include <linux/version.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)

#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#endif

#include <linux/fs.h>
#include <linux/serial.h>

#include <asm/atomic.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/ioctls.h>

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>

#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/smp_lock.h>

#include <linux/poll.h>
#include <linux/serial_reg.h>
#include <linux/miscdevice.h>
#include <linux/delay.h>
#include <linux/string.h>
#include <linux/slab.h>

#include "atarisio.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,23)
#define irqreturn_t void
#define IRQ_RETVAL(foo)
#endif

/* debug levels: */
#define DEBUG_STANDARD 1
#define DEBUG_NOISY 2
#define DEBUG_VERY_NOISY 3

/*
#define ATARISIO_DEBUG_TIMING
*/

/*
 * if ATARISIO_EARLY_NOTIFICATION is defined, poll will indicate a
 * pending command frame at the very start of the command frame,
 * not when it's completely received
 */
/*
#define ATARISIO_EARLY_NOTIFICATION
*/

#define PRINTK(x...) printk(NAME ": " x)

#define IRQ_PRINTK(level, x...) \
	if (debug_irq >= level) { \
		PRINTK(x); \
	}

#define DBG_PRINTK(level, x...) \
	if (debug >= level) { \
		PRINTK(x); \
	}

#ifdef ATARISIO_DEBUG_TIMING
#define PRINT_TIMESTAMP(msg) \
	do { \
		struct timeval tv; \
		do_gettimeofday(&tv); \
		printk(NAME ": %s (%lu:%lu)\n", msg, tv.tv_sec, tv.tv_usec);\
	} while (0)
#else
#define PRINT_TIMESTAMP(msg) do { } while(0)
#endif


/*
#define ATARISIO_DEBUG_TIMEOUTS
#define ATARISIO_DEBUG_TIMEOUT_WAIT (HZ/2)
*/

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
#ifdef MODULE_AUTHOR
MODULE_AUTHOR("Matthias Reichl <hias@horus.com>");
#endif
#ifdef MODULE_DESCRIPTION
MODULE_DESCRIPTION("Serial Atari 8bit SIO driver");
#endif

/* 
 * currently we use major 10, minor 240, which is a reserved local
 * number within the miscdevice block
 */

#define ATARISIO_DEFAULT_MINOR 240
#define NAME "atarisio"

/*
 * module parameters:
 * io:  the base address of the 16550
 * irq: interrupt number
 */
static char* port = 0;
static char* saved_port = 0;
static int io = 0;
static int irq = 0;
static int minor = ATARISIO_DEFAULT_MINOR;
static int debug = 0;
static int debug_irq = 0;

MODULE_PARM(port,"s");
MODULE_PARM(io,"i");
MODULE_PARM(irq,"i");
MODULE_PARM(minor,"i");
MODULE_PARM(debug,"i");
MODULE_PARM(debug_irq,"i");

#ifdef MODULE_PARM_DESC
MODULE_PARM_DESC(port,"serial port (eg /dev/ttyS0, default: use supplied io/irq values)");
MODULE_PARM_DESC(io,"io address of 16550 UART (eg 0x3f8)");
MODULE_PARM_DESC(irq,"irq of 16550 UART (eg 4)");
MODULE_PARM_DESC(minor,"minor device number (default: 240)");
MODULE_PARM_DESC(debug,"debug level (default: 0)");
MODULE_PARM_DESC(debug_irq,"interrupt debug level (default: 0)");
#endif

/*
 * constants of the SIO protocol
 */

#define MAX_COMMAND_FRAME_RETRIES 14
#define COMMAND_FRAME_ACK_CHAR 0x41
#define COMMAND_FRAME_NAK_CHAR 0x4e

#define OPERATION_COMPLETE_CHAR 0x43
#define OPERATION_ERROR_CHAR 0x45
#define DATA_FRAME_ACK_CHAR 0x41
#define DATA_FRAME_NAK_CHAR 0x4e


/*
 * delays, according to the SIO specs, in uSecs
 */

#define DELAY_T0 900
#define DELAY_T1 850
#define DELAY_T2_MIN 10
#define DELAY_T2_MAX 20000
#define DELAY_T3_MIN 1000
#define DELAY_T3_MAX 1600
#define DELAY_T4_MIN 450
#define DELAY_T4_MAX 16000

/* the QMEG OS needs at least 300usec delay between ACK and  complete */
/* #define DELAY_T5_MIN 250 */
#define DELAY_T5_MIN 300
#define DELAY_T5_MIN_SLOW 1000

/* this one is in mSecs, one jiffy (10 mSec on i386) is minimum,
 * try higher values for slower machines. The default of 50 mSecs
 * should be sufficient for most PCs
 */
#define DELAY_RECEIVE_HEADROOM 50
#define DELAY_SEND_HEADROOM 50

#define MODE_1050_2_PC 0
#define MODE_SIOSERVER 1

/* lock for UART hardware registers */
spinlock_t uart_lock = SPIN_LOCK_UNLOCKED;

static volatile int current_mode = MODE_1050_2_PC;

/*
 * helper functions to access the 16550
 */
static inline void serial_out(unsigned int offset, u8 value)
{
	if (offset >=8) {
		DBG_PRINTK(DEBUG_STANDARD, "illegal offset in serial_out\n");
	} else {
		outb(value, io+offset);
	}
}

static inline u8 serial_in(unsigned int offset)
{
	if (offset >=8) {
		DBG_PRINTK(DEBUG_STANDARD, "illegal offset in serial_in\n");
		return 0;
	} else {
		return inb(io+offset);
	}
}

/*
 * IO buffers for the intterupt handlers
 */

#define IOBUF_LENGTH 8200
/*#define IOBUF_LENGTH 300*/

#define MAX_SIO_DATA_LENGTH (IOBUF_LENGTH-2)
/* circular buffers can hold IOBUF_LENGTH-1 chars, reserve one further
 * char for the checksum
 */

static volatile struct rx_buf_struct {
	unsigned char buf[IOBUF_LENGTH];
	unsigned int head;
	unsigned int tail;
	int wakeup_len;
} rx_buf;

spinlock_t rx_lock = SPIN_LOCK_UNLOCKED;

static volatile struct tx_buf_struct {
	unsigned char buf[IOBUF_LENGTH];
	unsigned int head;
	unsigned int tail;
} tx_buf;

spinlock_t tx_lock = SPIN_LOCK_UNLOCKED;

static unsigned int current_cmdframe_serial_number;

static volatile struct cmdframe_buf_struct {
	unsigned char buf[5];
	unsigned int is_valid;
	unsigned int serial_number;
	unsigned int receiving;
	unsigned int pos;
	unsigned int error_counter;
	unsigned long long start_reception_time; /* in usec */
	unsigned long long end_reception_time; /* in usec */
	size_t missed_count;
} cmdframe_buf;

static unsigned long long get_timestamp(void)
{
	struct timeval tv;
	do_gettimeofday(&tv);
	return tv.tv_sec*1000000 + tv.tv_usec;
}

spinlock_t cmdframe_lock = SPIN_LOCK_UNLOCKED;

/*
 * wait queues
 */

wait_queue_head_t rx_queue;
wait_queue_head_t tx_queue;
wait_queue_head_t cmdframe_queue;

#ifdef ATARISIO_EARLY_NOTIFICATION
wait_queue_head_t early_cmdframe_queue;
#endif

static inline void reset_rx_buf(void)
{
	unsigned long flags;
	spin_lock_irqsave(&rx_lock, flags);

	if (rx_buf.head!=rx_buf.tail) {
		DBG_PRINTK(DEBUG_STANDARD, "reset_rx_buf: flushing %d characters\n",
			(rx_buf.head+IOBUF_LENGTH - rx_buf.tail) % IOBUF_LENGTH);
	}

	rx_buf.head=rx_buf.tail=0;
	rx_buf.wakeup_len=-1;

	spin_unlock_irqrestore(&rx_lock, flags);
}

static inline void reset_tx_buf(void)
{
	unsigned long flags;
	spin_lock_irqsave(&tx_lock, flags);

	if (tx_buf.head!=tx_buf.tail) {
		DBG_PRINTK(DEBUG_STANDARD, "reset_tx_buf: flushing %d characters\n",
			(tx_buf.head+IOBUF_LENGTH - tx_buf.tail) % IOBUF_LENGTH);
	}

	tx_buf.head=tx_buf.tail=0;

	spin_unlock_irqrestore(&tx_lock, flags);
}

static inline void reset_cmdframe_buf(int do_lock)
{
	unsigned long flags = 0;
	if (do_lock) {
		spin_lock_irqsave(&cmdframe_lock, flags);
	}

	cmdframe_buf.is_valid = 0;
	cmdframe_buf.serial_number=0;
	cmdframe_buf.receiving = 0;
	cmdframe_buf.pos = 0;
	cmdframe_buf.error_counter=0;
	cmdframe_buf.start_reception_time=0;
	cmdframe_buf.end_reception_time=0;
	cmdframe_buf.missed_count=0;

	if (do_lock) {
		spin_unlock_irqrestore(&cmdframe_lock, flags);
	}
}

/*
 * configuration of the serial port
 */

static volatile struct serial_config_struct {
	unsigned int baudrate;
	unsigned char IER;
	unsigned char MCR;
	unsigned char LCR;
	unsigned char FCR;
	unsigned long baud_base;
} serial_config;

spinlock_t serial_config_lock = SPIN_LOCK_UNLOCKED;


/* 
 * configuration for SIOSERVER mode
 */

static unsigned char sioserver_command_line = UART_MSR_RI;
static unsigned char sioserver_command_line_delta = UART_MSR_TERI;

/* 
 * configuration for 1050-2-PC mode
 */

/*
 * The command line has to be set to LOW during the
 * transmission of a command frame and HI otherwise
 */
static unsigned char command_line_mask = ~UART_MCR_RTS;
static unsigned char command_line_low = UART_MCR_RTS;
static unsigned char command_line_high = 0;

static unsigned int default_baudrate = ATARISIO_STANDARD_BAUDRATE;
static int do_autobaud = 0;
static int add_highspeedpause = 0;

static unsigned char standard_lcr = UART_LCR_WLEN8;
static unsigned char slow_lcr = UART_LCR_WLEN8 | UART_LCR_STOP | UART_LCR_PARITY | UART_LCR_SPAR;

/*
 * indicate the start of a command frame
 */
static inline void set_command_line(void)
{
	unsigned long flags;
	spin_lock_irqsave(&uart_lock, flags);

	serial_config.MCR = (serial_config.MCR & command_line_mask) | command_line_low;
	serial_out(UART_MCR, serial_config.MCR);

	spin_unlock_irqrestore(&uart_lock, flags);
}

/*
 * indicate the end of a command frame
 */
static inline void clear_command_line(int lock_uart)
{
	unsigned long flags = 0;

	if (lock_uart) {
		spin_lock_irqsave(&uart_lock, flags);
	}

	serial_config.MCR = (serial_config.MCR & command_line_mask) | command_line_high;
	serial_out(UART_MCR, serial_config.MCR);

	if (lock_uart) {
		spin_unlock_irqrestore(&uart_lock, flags);
	}
}

static void set_lcr(unsigned char lcr, int do_locks)
{
	unsigned long flags=0;

	if (do_locks) {
		spin_lock_irqsave(&uart_lock, flags);
		spin_lock(&serial_config_lock);
	}

	serial_config.LCR = lcr;
	serial_out(UART_LCR, serial_config.LCR);

	if (do_locks) {
		spin_unlock(&serial_config_lock);
		spin_unlock_irqrestore(&uart_lock, flags);
	}
}

static int set_baudrate(unsigned int baudrate, int do_locks)
{
	unsigned long flags=0;
	unsigned int divisor;

	if (do_locks) {
		spin_lock_irqsave(&uart_lock, flags);
		spin_lock(&serial_config_lock);
	}

	serial_config.baudrate = baudrate;
	divisor = serial_config.baud_base / baudrate;

	serial_out(UART_LCR, serial_config.LCR | UART_LCR_DLAB);
	serial_out(UART_DLL, divisor & 0xff);
	serial_out(UART_DLM, divisor >> 8);
	serial_out(UART_LCR, serial_config.LCR);

	if (do_locks) {
		spin_unlock(&serial_config_lock);
		spin_unlock_irqrestore(&uart_lock, flags);
	}

	return 0;
}


static inline unsigned char calculate_checksum(const unsigned char* circ_buf, int start, int len, int total_len)
{
	unsigned int chksum=0;
	int i;
	for (i=0;i<len;i++) {
		chksum += circ_buf[(start+i) % total_len];
		if (chksum >= 0x100) {
			chksum = (chksum & 0xff) + 1;
		}
	}
	return chksum;
}

static inline void reset_fifos(void)
{
	IRQ_PRINTK(DEBUG_NOISY, "resetting 16550 fifos\n");

	serial_out(UART_FCR, 0);
	serial_out(UART_FCR, serial_config.FCR | UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT);
}

static inline void receive_chars(void)
{
	unsigned char c,lsr;
	int do_wakeup=0;
	int first = 1;

	spin_lock(&rx_lock);
	spin_lock(&cmdframe_lock);
	
	while ( (lsr=serial_in(UART_LSR)) & UART_LSR_DR ) {
		if (first && (serial_config.LCR != standard_lcr) ) {
			IRQ_PRINTK(DEBUG_STANDARD, "wrong LCR in receive chars!\n");
			spin_lock(&serial_config_lock);
			set_lcr(standard_lcr, 0);
			spin_unlock(&serial_config_lock);
		}
		first = 0;
		if (lsr & UART_LSR_OE) {
			IRQ_PRINTK(DEBUG_STANDARD, "overrun error\n");
		}
		c = serial_in(UART_RX);

		if (lsr & UART_LSR_FE) {
			IRQ_PRINTK(DEBUG_NOISY, "got framing error\n");
			if (cmdframe_buf.receiving) {
				cmdframe_buf.error_counter++;
			}
		} else {
			if (cmdframe_buf.receiving) {
				if (cmdframe_buf.pos<5) {
					cmdframe_buf.buf[cmdframe_buf.pos] = c;
				} else {
					cmdframe_buf.error_counter++;
				}
				cmdframe_buf.pos++;
			} else {
				if ( (rx_buf.head+1) % IOBUF_LENGTH != rx_buf.tail ) {
					rx_buf.buf[rx_buf.head] = c;
					rx_buf.head = (rx_buf.head+1) % IOBUF_LENGTH;
					if ( (rx_buf.head + IOBUF_LENGTH - rx_buf.tail) % IOBUF_LENGTH == (unsigned int) rx_buf.wakeup_len) {
						do_wakeup=1;
					}
					IRQ_PRINTK(DEBUG_VERY_NOISY, "received character 0x%02x\n",c);
				} else {
					IRQ_PRINTK(DEBUG_VERY_NOISY, "sinking character 0x%02x\n",c);
				}
			}
		}
	}
	spin_unlock(&cmdframe_lock);
	spin_unlock(&rx_lock);
	if ( do_wakeup ) {
		wake_up(&rx_queue);
	}
}

static inline void send_chars(void)
{
	int do_wakeup=0;
	int count = 16; /* FIFO can hold 16 chars at maximum */
	unsigned char lsr;

	lsr = serial_in(UART_LSR);
	if (lsr & UART_LSR_OE) {
		IRQ_PRINTK(DEBUG_STANDARD, "overrun error\n");
	}
	if (lsr & UART_LSR_THRE) {
		spin_lock(&tx_lock);
		while ( (count > 0) && (tx_buf.head != tx_buf.tail) ) {
			IRQ_PRINTK(DEBUG_VERY_NOISY, "transmit char 0x%02x\n",tx_buf.buf[tx_buf.head]);

			serial_out(UART_TX, tx_buf.buf[tx_buf.tail]);
			tx_buf.tail = (tx_buf.tail+1) % IOBUF_LENGTH;
			count--;
		};
		if ( tx_buf.head == tx_buf.tail ) {
			/* end of TX-buffer reached, disable further TX-interrupts */
			spin_lock(&serial_config_lock);

			serial_config.IER &= ~UART_IER_THRI;
			serial_out(UART_IER, serial_config.IER);

			spin_unlock(&serial_config_lock);
			do_wakeup=1;
		}
		spin_unlock(&tx_lock);
		if (do_wakeup) {
			wake_up(&tx_queue);
		}
	}
}

static inline void try_switchbaud(void)
{
	unsigned int baud;

	spin_lock(&serial_config_lock);

        baud= serial_config.baudrate;

	if (do_autobaud) {
		if (baud == ATARISIO_STANDARD_BAUDRATE) {
			baud = ATARISIO_HIGHSPEED_BAUDRATE;
		/*
		} else if (baud == 57600) {
			baud = 38400;
		*/
		} else {
			baud = ATARISIO_STANDARD_BAUDRATE;
		}

		IRQ_PRINTK(DEBUG_STANDARD, "switching to %d baud\n",baud);
	}

	set_baudrate(baud, 0);

	spin_unlock(&serial_config_lock);
}

static inline void check_modem_lines_before_receive(unsigned char old_msr, unsigned char new_msr)
{
	if (current_mode == MODE_SIOSERVER) {
		if ( (new_msr & sioserver_command_line) != (old_msr & sioserver_command_line)) {
			IRQ_PRINTK(DEBUG_VERY_NOISY, "msr changed from 0x%02x to 0x%02x [%ld]\n",old_msr, new_msr, jiffies);
			if (new_msr & sioserver_command_line) {
				PRINT_TIMESTAMP("start of command frame");
				/* start of a new command frame */
				spin_lock(&cmdframe_lock);
				if (cmdframe_buf.is_valid) {
					IRQ_PRINTK(DEBUG_STANDARD, "invalidating command frame (detected new frame) %d\n", 
						cmdframe_buf.missed_count);
					cmdframe_buf.missed_count++;
				}
				if (cmdframe_buf.receiving) {
					IRQ_PRINTK(DEBUG_STANDARD, "restarted reception of command frame (detected new frame)\n");
				}
				cmdframe_buf.receiving = 1;
				cmdframe_buf.start_reception_time = get_timestamp();
				cmdframe_buf.pos = 0;
				cmdframe_buf.is_valid = 0;
				cmdframe_buf.error_counter = 0;
				cmdframe_buf.serial_number++;

				spin_unlock(&cmdframe_lock);

#ifdef ATARISIO_EARLY_NOTIFICATION
				wake_up(&early_cmdframe_queue);
#endif
			}
		}
	}
}

#define DEBUG_PRINT_CMDFRAME_BUF \
	"cmdframe_buf: %02x %02x %02x %02x %02x err: %d pos=%d\n",\
	cmdframe_buf.buf[0],\
	cmdframe_buf.buf[1],\
	cmdframe_buf.buf[2],\
	cmdframe_buf.buf[3],\
	cmdframe_buf.buf[4],\
	cmdframe_buf.error_counter,\
	cmdframe_buf.pos

static inline void check_modem_lines_after_receive(unsigned char old_msr, unsigned char new_msr)
{
	int do_wakeup = 0;
	int OK = 1;
	unsigned char checksum;

	if (current_mode == MODE_SIOSERVER) {
		/*
		 * check if the UART indicated a change on the command line from
		 * high to low. If yes, a complete (5 byte) command frame should
		 * have been received.
		 */
		if ( (new_msr & sioserver_command_line_delta) && (!(new_msr & sioserver_command_line)) ) {
			spin_lock(&rx_lock);
			spin_lock(&cmdframe_lock);


			if (rx_buf.tail != rx_buf.head) {
				IRQ_PRINTK(DEBUG_STANDARD, "flushing rx buffer (%d/%d)\n", rx_buf.head, rx_buf.tail);
			}
			if (tx_buf.tail != tx_buf.head) {
				IRQ_PRINTK(DEBUG_STANDARD, "flushing tx buffer (%d/%d)\n", tx_buf.head, tx_buf.tail);
			}

			rx_buf.tail = rx_buf.head; /* flush input buffer */
			tx_buf.tail = tx_buf.head; /* flush output buffer */

			if (cmdframe_buf.pos != 5) {
				if (cmdframe_buf.pos <=3 || cmdframe_buf.pos >=7) {
					cmdframe_buf.error_counter+=5;
				} else {
					cmdframe_buf.error_counter++;
				}
				IRQ_PRINTK(DEBUG_NOISY, "got %d chars for this command frame\n",
					cmdframe_buf.pos);
				OK = 0;
			}

			if (! cmdframe_buf.receiving) {
			 	IRQ_PRINTK(DEBUG_NOISY, "indicated end of command frame, but I'm not currently receiving\n");
				OK = 0;
			}

			if (cmdframe_buf.error_counter>0) {
				OK = 0;
			}

			if (OK) {
				IRQ_PRINTK(DEBUG_NOISY, DEBUG_PRINT_CMDFRAME_BUF);

				checksum = calculate_checksum((unsigned char*)cmdframe_buf.buf, 0, 4, 5);

				if (cmdframe_buf.buf[4] == checksum) {
					PRINT_TIMESTAMP("found command frame");
					IRQ_PRINTK(DEBUG_NOISY, "found command frame [%ld]\n", jiffies);
					do_wakeup = 1;

					cmdframe_buf.is_valid = 1;
					cmdframe_buf.end_reception_time = get_timestamp();
					cmdframe_buf.error_counter = 0;

				} else {
					IRQ_PRINTK(DEBUG_STANDARD, "command frame checksum error [%ld]\n",jiffies);
					IRQ_PRINTK(DEBUG_NOISY, DEBUG_PRINT_CMDFRAME_BUF);

					cmdframe_buf.is_valid = 0;
					cmdframe_buf.error_counter++;
					cmdframe_buf.pos=0;
				}
			} else {
				IRQ_PRINTK(DEBUG_STANDARD, "errors in command frame reception\n");
				IRQ_PRINTK(DEBUG_NOISY, DEBUG_PRINT_CMDFRAME_BUF);
			}

			if (cmdframe_buf.error_counter >= 3) {
				IRQ_PRINTK(DEBUG_NOISY, "error counter exceeded limit [%d]\n",cmdframe_buf.error_counter);
				try_switchbaud();

				/* some 16550 chips get confused if we switch the baudrate too
				 * fast. clearing the FIFOs solves this problem, but don't ask
				 * me why...
				 */
				reset_fifos();

			}
			cmdframe_buf.receiving = 0;

			spin_unlock(&cmdframe_lock);
			spin_unlock(&rx_lock);

			if (do_wakeup) {
				PRINT_TIMESTAMP("waking up cmdframe_queue");
				wake_up(&cmdframe_queue);
			}
		}
	}
}

/*
 * the interrupt handler
 */

static irqreturn_t atarisio_interrupt(int irq, void* dev_id, struct pt_regs* regs)
{
	unsigned char iir,msr;
	static unsigned char last_msr=0;
	unsigned int handled = 0;

	spin_lock(&uart_lock);

	while (! ((iir=serial_in(UART_IIR)) & UART_IIR_NO_INT)) {
		handled = 1;

		msr = serial_in(UART_MSR);

		check_modem_lines_before_receive(last_msr, msr);

		receive_chars();

		check_modem_lines_after_receive(last_msr, msr);

		send_chars();

		last_msr = msr;
	}

	spin_unlock(&uart_lock);

	return IRQ_RETVAL(handled);
}

// check if a new command frame was received, before sending
// a command ACK/NAK
static inline int check_new_command_frame(void)
{
	if (cmdframe_buf.serial_number != current_cmdframe_serial_number) {
		DBG_PRINTK(DEBUG_STANDARD, "new command frame has arrived...\n");
		return -EATARISIO_COMMAND_TIMEOUT;
	} else {
		return 0;
	}
}

static inline int check_command_frame_time(int do_lock)
{
	unsigned long flags = 0;
	unsigned long long current_time;
	int ret = 0;
	unsigned int do_delay = 0;

	if (do_lock) {
		spin_lock_irqsave(&cmdframe_lock, flags);
	}
	current_time = get_timestamp();

	if (current_time > cmdframe_buf.end_reception_time+10000 ) {
		DBG_PRINTK(DEBUG_STANDARD, "command frame is too old (%lu usecs)\n",
				(unsigned long) (current_time - cmdframe_buf.end_reception_time));
		ret = -EATARISIO_COMMAND_TIMEOUT;
	} else {
		DBG_PRINTK(DEBUG_NOISY, "command frame age is OK (%lu usecs)\n",
				(unsigned long) (current_time - cmdframe_buf.end_reception_time));
		if (do_lock && (current_time - cmdframe_buf.end_reception_time < DELAY_T2_MIN)) {
			do_delay = DELAY_T2_MIN-(current_time - cmdframe_buf.end_reception_time);
		}
	}
	if (do_lock) {
		spin_unlock_irqrestore(&cmdframe_lock, flags);
	}
	if (do_delay>0) {
		udelay(do_delay);
	}
	return ret;
}

static inline void initiate_send(void)
{
	unsigned long flags;
	DBG_PRINTK(DEBUG_VERY_NOISY, "initiate_send: tx-head = %d tx-tail = %d\n", tx_buf.head, tx_buf.tail);

	if ( (tx_buf.head != tx_buf.tail)
	     && !(serial_config.IER & UART_IER_THRI) ) {

		DBG_PRINTK(DEBUG_VERY_NOISY, "enabling TX-interrupt\n");

		spin_lock_irqsave(&uart_lock, flags);
		spin_lock(&serial_config_lock);

		serial_config.IER |= UART_IER_THRI;
		serial_out(UART_IER, serial_config.IER);

		spin_unlock(&serial_config_lock);
		spin_unlock_irqrestore(&uart_lock, flags);
	}
}

static inline int wait_send(size_t len)
{
	wait_queue_t wait;
	signed long timeout_jiffies;
	signed long expire;
	unsigned int max_jiffies;
	int timeout;

	timeout = 10*len*1000 / serial_config.baudrate;
	timeout += DELAY_SEND_HEADROOM;

	timeout_jiffies = timeout*HZ/1000;
	if (timeout_jiffies <= 0) {
		timeout_jiffies = 1;
	}

	expire = timeout_jiffies;
	max_jiffies = jiffies + timeout_jiffies;

	/*
	 * first wait for the interrupt notification, telling us
	 * that the transmit buffer is empty
	 */
	init_waitqueue_entry(&wait, current);

	add_wait_queue(&tx_queue, &wait);
	while (1) {
		current->state = TASK_INTERRUPTIBLE;
		if (tx_buf.head == tx_buf.tail) {
			break;
		}
		expire = schedule_timeout(expire);
		if (expire==0) {
			break;
		}
		if (signal_pending(current)) {
			current->state=TASK_RUNNING;
			remove_wait_queue(&tx_queue, &wait);
			return -EINTR;
		}
	}
	current->state=TASK_RUNNING;
	remove_wait_queue(&tx_queue, &wait);

	if ( tx_buf.head != tx_buf.tail ) {
		DBG_PRINTK(DEBUG_STANDARD, "timeout expired in wait_send\n");
		return -EATARISIO_COMMAND_TIMEOUT;
	}

	/*
	 * now wait for the 16550 FIFO and transmitter to empty
	 */
	while ((jiffies < max_jiffies) && !(serial_in(UART_LSR) & UART_LSR_TEMT)) {
	}
	if (jiffies >= max_jiffies) {
		DBG_PRINTK(DEBUG_STANDARD, "timeout expired in wait_send / TEMT\n");
		return -EATARISIO_COMMAND_TIMEOUT;
	} else {
		return 0;
	}
}

static inline int wait_receive(int len, int additional_timeout)
{
	wait_queue_t wait;

	int no_received_chars=0;
	int timeout;
	signed long timeout_jiffies;
	signed long expire;
	signed long jiffies_start = jiffies;

	timeout = 10*len*1000 / serial_config.baudrate;
	timeout += additional_timeout + DELAY_RECEIVE_HEADROOM;

	timeout_jiffies = timeout*HZ/1000;
	if (timeout_jiffies <= 0) {
		timeout_jiffies = 1;
	}

	expire = timeout_jiffies;

	init_waitqueue_entry(&wait, current);

	rx_buf.wakeup_len = len;
	add_wait_queue(&rx_queue, &wait);
	while (1) {
		current->state = TASK_INTERRUPTIBLE;
		if ( (no_received_chars=(rx_buf.head+IOBUF_LENGTH - rx_buf.tail) % IOBUF_LENGTH) >=len) {
			break;
		}
		expire = schedule_timeout(expire);
		if (expire==0) {
			break;
		}
		if (signal_pending(current)) {
			current->state=TASK_RUNNING;
			remove_wait_queue(&rx_queue, &wait);
			return -EINTR;
		}
	}
	current->state=TASK_RUNNING;
	remove_wait_queue(&rx_queue, &wait);

#ifdef ATARISIO_DEBUG_TIMEOUTS
	no_received_chars=(rx_buf.head+IOBUF_LENGTH - rx_buf.tail) % IOBUF_LENGTH;
	if (no_received_chars != len) {
		check_new_command_frame();
		DBG_PRINTK(DEBUG_STANDARD, "timeout in wait_receive [wanted %d got %d time %ld], waiting again\n",
				len, no_received_chars, jiffies-jiffies_start);
		rx_buf.wakeup_len = len;
		expire = ATARISIO_DEBUG_TIMEOUT_WAIT;
		add_wait_queue(&rx_queue, &wait);
		while(1) {
			current->state = TASK_INTERRUPTIBLE;
			if ((no_received_chars=(rx_buf.head+IOBUF_LENGTH - rx_buf.tail) % IOBUF_LENGTH) >=len) {
				break;
			}
			expire = schedule_timeout(expire);
			if (expire==0) {
				break;
			}
			if (signal_pending(current)) {
				current->state=TASK_RUNNING;
				remove_wait_queue(&rx_queue, &wait);
				return -EINTR;
			}
		}
		current->state=TASK_RUNNING;
		remove_wait_queue(&rx_queue, &wait);
	}
#endif
	rx_buf.wakeup_len=-1;
	no_received_chars=(rx_buf.head+IOBUF_LENGTH - rx_buf.tail) % IOBUF_LENGTH;
	if (no_received_chars < len) {
		DBG_PRINTK(DEBUG_STANDARD, "timeout in wait_receive [wanted %d got %d time %ld]\n",
				len, no_received_chars, jiffies - jiffies_start);
	}
	if (no_received_chars > len) {
		DBG_PRINTK(DEBUG_NOISY, "received more bytes than expected [expected %d got %d]\n",
				len, no_received_chars);
	}
	return no_received_chars;
}

static inline int send_command_frame(SIO_parameters* params)
{
	int retry = 0;
	int last_err = 0;
	unsigned long flags;
	int i;
	unsigned char c;
	int w;

	unsigned char cmd_frame[5];

	cmd_frame[0] = params->device_id;
	cmd_frame[1] = params->command;
	cmd_frame[2] = params->aux1;
	cmd_frame[3] = params->aux2;
	cmd_frame[4] = calculate_checksum(cmd_frame, 0, 4, 5);

	while (retry < MAX_COMMAND_FRAME_RETRIES) {
		DBG_PRINTK(DEBUG_NOISY, "initiating command frame [%ld]\n",jiffies);

		rx_buf.tail = rx_buf.head; /* clear rx-buffer */

		set_command_line();
		udelay(DELAY_T0);

		spin_lock_irqsave(&tx_lock, flags);
		if (tx_buf.head != tx_buf.tail) {
			DBG_PRINTK(DEBUG_NOISY, "clearing tx bufffer\n");
			tx_buf.head = tx_buf.tail;
		}
		for (i=0;i<5;i++) {
			tx_buf.buf[tx_buf.head+i] = cmd_frame[i];
		}
		tx_buf.head = (tx_buf.head+5) % IOBUF_LENGTH;

		initiate_send();
		spin_unlock_irqrestore(&tx_lock, flags);
		if ((w=wait_send(5))) {
			if (w == -EATARISIO_COMMAND_TIMEOUT) {
				last_err = w;
				goto again;
			} else {
				DBG_PRINTK(DEBUG_STANDARD, "sending command frame returned %d\n",w);
				return w;
			}
		}

		udelay(DELAY_T1);
		clear_command_line(1);
		udelay(DELAY_T2_MIN);

		PRINT_TIMESTAMP("begin wait for command ACK");
		w=wait_receive(1, DELAY_T2_MAX/1000);
		PRINT_TIMESTAMP("end wait for command ACK");

		if (w < 1) {
			if (w < 0) {
				DBG_PRINTK(DEBUG_STANDARD, "wait_receive returned %d\n",w);
				return w;
			} else {
				DBG_PRINTK(DEBUG_STANDARD, "waiting for command frame ACK timed out [%ld]\n", jiffies);

				last_err = -EATARISIO_COMMAND_TIMEOUT;
				goto again;
			}
		}
		c = rx_buf.buf[rx_buf.tail];
		rx_buf.tail = (rx_buf.tail+1) % IOBUF_LENGTH;
		if (c==COMMAND_FRAME_ACK_CHAR) {
			break;
		} else if (c==COMMAND_FRAME_NAK_CHAR) {
			DBG_PRINTK(DEBUG_STANDARD, "got command frame NAK\n");
			last_err = -EATARISIO_COMMAND_NAK;
		} else {
			DBG_PRINTK(DEBUG_STANDARD, "illegal response to command frame: 0x%02x\n",c);
		}
		if (retry+1 < MAX_COMMAND_FRAME_RETRIES) {
			DBG_PRINTK(DEBUG_NOISY, "retrying command frame\n");
		}

again:
		clear_command_line(1);
		{
			signed long t = HZ / 5;
			current->state=TASK_INTERRUPTIBLE;
			while (t) {
				t = schedule_timeout(t);
				if (signal_pending(current)) {
					current->state=TASK_RUNNING;
					return -EINTR;
				}
			}
			current->state=TASK_RUNNING;
		}
		retry++;
	}
		
	if (retry == MAX_COMMAND_FRAME_RETRIES) {
		DBG_PRINTK(DEBUG_STANDARD, "request timed out - no ACK to command frame\n");

		if (last_err) {
			return last_err;
		} else {
			return -EATARTSIO_UNKNOWN_ERROR;
		}
	} else {
		DBG_PRINTK(DEBUG_NOISY, "got ACK for command frame - ready to proceed\n");
		return 0;
	}
}

static int setup_send_data_frame(size_t data_length, unsigned char* user_buffer)
{
	unsigned long flags;
	unsigned int len, remain;
	unsigned char checksum;

	if (data_length == 0) {
		return -EINVAL;
	}

	spin_lock_irqsave(&tx_lock, flags);
	tx_buf.head = tx_buf.tail;

	len = IOBUF_LENGTH - tx_buf.head;
	if (len > data_length) {
		len = data_length;
	}
	if ( copy_from_user((unsigned char*) (&(tx_buf.buf[tx_buf.head])), 
			user_buffer,
			len) ) {
		return -EFAULT;
	}
	remain = data_length - len;
	if (remain>0) {
		if ( copy_from_user((unsigned char*) tx_buf.buf,
				user_buffer+len,
				remain) ) {
			return -EFAULT;
		}
	}

	checksum = calculate_checksum((unsigned char*)(tx_buf.buf),
		       	tx_buf.head, data_length, IOBUF_LENGTH);

	tx_buf.buf[(tx_buf.head+data_length) % IOBUF_LENGTH] = checksum;

	tx_buf.head = (tx_buf.head + data_length + 1) % IOBUF_LENGTH;
	spin_unlock_irqrestore(&tx_lock, flags);

	return 0;
}

static int copy_received_data_to_user(size_t data_length, unsigned char* user_buffer)
{
	unsigned int len, remain;
	if (data_length == 0) {
		return -EINVAL;
	}
	/* if we received any bytes, copy the to the user buffer, even
	 * if we ran into a timeout and got less than we wanted
	 */
	len = IOBUF_LENGTH - rx_buf.tail;
	if (len > data_length) {
		len = data_length;
	}
	if ( copy_to_user(user_buffer,
			 (unsigned char*) (&(rx_buf.buf[rx_buf.tail])),
			 len) ) {
		return -EFAULT;
	}
	remain = data_length - len;
	if (remain>0) {
		if ( copy_to_user(user_buffer+len,
				 (unsigned char*) rx_buf.buf,
				 remain) ) {
			return -EFAULT;
		}
	}
	return 0;
}

static int perform_sio_receive_data(SIO_parameters * sio_params)
{
	int received_len, copy_len;
        int ret=0,w;
	unsigned char c, checksum;
	/*
	 * read data from disk
	 */
	if ((w=wait_receive(1, sio_params->timeout*1000)) < 1) {
		if (w < 0) {
			DBG_PRINTK(DEBUG_STANDARD, "wait_receive returned %d\n",w);
			return w;
		} else {
			DBG_PRINTK(DEBUG_STANDARD, "command complete timed out\n");
			return -EATARISIO_COMMAND_TIMEOUT;
		}
	}
	ret = 0;
	c = rx_buf.buf[rx_buf.tail];
	rx_buf.tail = (rx_buf.tail+1) % IOBUF_LENGTH;

	if (c != OPERATION_COMPLETE_CHAR) {
		if (c == COMMAND_FRAME_ACK_CHAR) {
			DBG_PRINTK(DEBUG_STANDARD, "got command ACK instead of command complete\n");
		} else {
			if (c != OPERATION_ERROR_CHAR) {
				DBG_PRINTK(DEBUG_STANDARD, "wanted command complete, got: 0x%02x\n",c);
			} else {
				DBG_PRINTK(DEBUG_STANDARD, "got sio command error\n");
			}
			ret = -EATARISIO_COMMAND_COMPLETE_ERROR;
		}
	}

	if (sio_params->data_length) {
		/* receive data block and checksum*/
		received_len = wait_receive(sio_params->data_length+1, 0);

		if (received_len < 0) {
			DBG_PRINTK(DEBUG_STANDARD, "receiving data block returned %d\n",received_len);
			return received_len;
		}

		copy_len = sio_params->data_length;
		if (received_len < copy_len) {
			copy_len = received_len;
		}
		if (copy_len > 0) {
			if ( (w=copy_received_data_to_user(copy_len, sio_params->data_buffer)) ) {
				return w;
			}
		}

	        if ((unsigned int) received_len < sio_params->data_length+1) {
			DBG_PRINTK(DEBUG_STANDARD, "receive data frame timed out [wanted %d got %d]\n",
					sio_params->data_length+1, received_len);
			return -EATARISIO_COMMAND_TIMEOUT;
		}
		checksum = calculate_checksum((unsigned char*)(rx_buf.buf),
			       	rx_buf.tail, sio_params->data_length, IOBUF_LENGTH);

		if (rx_buf.buf[(rx_buf.tail+sio_params->data_length) % IOBUF_LENGTH ]
			!= checksum) {
			DBG_PRINTK(DEBUG_STANDARD, "checksum error : 0x%02x != 0x%02x\n",
				checksum,
				rx_buf.buf[(rx_buf.tail+sio_params->data_length) % IOBUF_LENGTH]);
			if (! ret) {
				ret = -EATARISIO_CHECKSUM_ERROR;
			}
		}


		rx_buf.tail = (rx_buf.tail + sio_params->data_length + 1 ) % IOBUF_LENGTH;

	}

	if (rx_buf.head != rx_buf.tail) {
		DBG_PRINTK(DEBUG_NOISY, "detected excess characters\n");
	}
	return ret;
}

static int perform_sio_send_data(SIO_parameters * sio_params)
{
        int ret=0,w;
	unsigned char c;

	/*
	 * read data from disk
	 */
	udelay(DELAY_T3_MIN);


	if (sio_params->data_length) {
		if ( (ret=setup_send_data_frame(sio_params->data_length, sio_params->data_buffer)) ) {
			return ret;
		}

		initiate_send();

		if ((w=wait_send(sio_params->data_length+1))) {
			DBG_PRINTK(DEBUG_STANDARD, "send data frame returned %d\n",w);
			return w;
		}

		/*
		 * wait for data frame ack
		 */

		if ((w=wait_receive(1, DELAY_T4_MAX/1000)) < 1) {
			if (w < 0) {
				DBG_PRINTK(DEBUG_STANDARD, "receive data frame ACK returned %d\n",w);
				return w;
			} else {
				DBG_PRINTK(DEBUG_STANDARD, "receive data frame ACK timed out\n");
				return -EATARISIO_COMMAND_TIMEOUT;
			}
		}
		c = rx_buf.buf[rx_buf.tail];
		rx_buf.tail = (rx_buf.tail + 1) % IOBUF_LENGTH;

		if (c != DATA_FRAME_ACK_CHAR) {
			if (c != DATA_FRAME_NAK_CHAR) {
				DBG_PRINTK(DEBUG_STANDARD, "wanted data ACK or ERROR, got: 0x%02x\n",c);
			}
			return -EATARISIO_DATA_NAK;
		}
	}

	/* 
	 * wait for command complete
	 */
	if ((w=wait_receive(1, sio_params->timeout*1000)) < 1) {
		if (w < 0) {
			DBG_PRINTK(DEBUG_STANDARD, "receive command complete returned %d\n",w);
			return w;
		} else {
			DBG_PRINTK(DEBUG_STANDARD, "receive command complete timed out\n");
			return -EATARISIO_COMMAND_TIMEOUT;
		}
	}
	c = rx_buf.buf[rx_buf.tail];
	rx_buf.tail = (rx_buf.tail + 1) % IOBUF_LENGTH;

	if (c != OPERATION_COMPLETE_CHAR) {
		if (c == COMMAND_FRAME_ACK_CHAR) {
			DBG_PRINTK(DEBUG_STANDARD, "got command ACK instead of command complete\n");
		} else {
			if (c != OPERATION_ERROR_CHAR) {
				DBG_PRINTK(DEBUG_STANDARD, "wanted command complete, got: 0x%02x\n",c);
			} else {
				DBG_PRINTK(DEBUG_STANDARD, "got sio command error\n");
			}
			ret = -EATARISIO_COMMAND_COMPLETE_ERROR;
		}
	}
	return ret;
}

static inline int perform_sio(SIO_parameters * user_sio_params)
{
	int len;
	int ret=0;
	SIO_parameters sio_params;

	len = sizeof(sio_params);
	if (copy_from_user(&sio_params, user_sio_params, sizeof(sio_params))) {
		DBG_PRINTK(DEBUG_STANDARD, "copy_from_user failed for SIO parameters\n");
		return -EFAULT;
	}

	if (sio_params.data_length >= MAX_SIO_DATA_LENGTH) {
		return -EATARISIO_ERROR_BLOCK_TOO_LONG;
	}

	DBG_PRINTK(DEBUG_VERY_NOISY, "performing SIO:\n");
	DBG_PRINTK(DEBUG_VERY_NOISY, "unit=%d cmd=%d dir=%d timeout=%d aux1=%d aux2=%d datalen=%d\n",
		sio_params.device_id, sio_params.command, sio_params.direction,
		sio_params.timeout, sio_params.aux1, sio_params.aux2, sio_params.data_length);

	if ((ret = send_command_frame(&sio_params))) {
		return ret;
	}

	if (sio_params.direction == 0) {
		ret = perform_sio_receive_data(&sio_params);
	} else {
		ret = perform_sio_send_data(&sio_params);
	}
	return ret;
}

static void set_1050_2_pc_mode(int do_locks)
{
	unsigned long flags=0;
	int have_to_wait;

	if (do_locks) {
		spin_lock_irqsave(&uart_lock, flags);
		spin_lock(&serial_config_lock);
	}

	have_to_wait = !(serial_config.MCR & UART_MCR_DTR);

	current_mode = MODE_1050_2_PC;
	command_line_mask = ~(UART_MCR_RTS | UART_MCR_DTR);
	command_line_low = UART_MCR_DTR | UART_MCR_RTS;
	command_line_high = UART_MCR_DTR;
	clear_command_line(0);

	set_lcr(standard_lcr, 0);
	do_autobaud = 0;
	set_baudrate(default_baudrate, 0);
	serial_config.IER &= ~UART_IER_MSI;
	serial_out(UART_IER, serial_config.IER);

	if (do_locks) {
		spin_unlock(&serial_config_lock);
		spin_unlock_irqrestore(&uart_lock, flags);
		if (have_to_wait) {
			schedule_timeout(HZ/2); /* wait 0.5 sec. for voltage supply to stabilize */
		}
	}

	reset_cmdframe_buf(1);
}

static void set_prosystem_mode(int do_locks)
{
	unsigned long flags = 0;
	int have_to_wait;

	if (do_locks) {
		spin_lock_irqsave(&uart_lock, flags);
		spin_lock(&serial_config_lock);
	}

	have_to_wait = !(serial_config.MCR & UART_MCR_RTS);

	current_mode = MODE_1050_2_PC;
	command_line_mask = ~(UART_MCR_RTS | UART_MCR_DTR);
	command_line_low = UART_MCR_RTS | UART_MCR_DTR;
	command_line_high = UART_MCR_RTS;
	clear_command_line(0);
	set_lcr(standard_lcr, 0);
	do_autobaud = 0;
	set_baudrate(default_baudrate, 0);
	serial_config.IER &= ~UART_IER_MSI;
	serial_out(UART_IER, serial_config.IER);

	if (do_locks) {
		spin_unlock(&serial_config_lock);
		spin_unlock_irqrestore(&uart_lock, flags);
		if (have_to_wait) {
			schedule_timeout(HZ/2); /* wait 0.5 sec. for voltage supply to stabilize */
		}
	}

	reset_cmdframe_buf(1);
}

static void set_sioserver_mode(int do_locks)
{
	unsigned long flags = 0;

	if (do_locks) {
		spin_lock_irqsave(&uart_lock, flags);
		spin_lock(&serial_config_lock);
	}

	current_mode = MODE_SIOSERVER;
	sioserver_command_line = UART_MSR_RI;
	sioserver_command_line_delta = UART_MSR_TERI;

	set_lcr(standard_lcr, 0);
	set_baudrate(default_baudrate, 0);
	serial_config.IER |= UART_IER_MSI;
	serial_out(UART_IER, serial_config.IER);

	if (do_locks) {
		spin_unlock(&serial_config_lock);
		spin_unlock_irqrestore(&uart_lock, flags);
	}

	reset_cmdframe_buf(1);
}

static void set_sioserver_mode_dsr(int do_locks)
{
	unsigned long flags = 0;

	if (do_locks) {
		spin_lock_irqsave(&uart_lock, flags);
		spin_lock(&serial_config_lock);
	}

	current_mode = MODE_SIOSERVER;
	sioserver_command_line = UART_MSR_DSR;
	sioserver_command_line_delta = UART_MSR_DDSR;

	set_lcr(standard_lcr, 0);
	set_baudrate(default_baudrate, 0);
	serial_config.IER |= UART_IER_MSI;
	serial_out(UART_IER, serial_config.IER);

	if (do_locks) {
		spin_unlock(&serial_config_lock);
		spin_unlock_irqrestore(&uart_lock, flags);
	}

	reset_cmdframe_buf(1);
}

static void set_sioserver_mode_cts(int do_locks)
{
	unsigned long flags = 0;

	if (do_locks) {
		spin_lock_irqsave(&uart_lock, flags);
		spin_lock(&serial_config_lock);
	}

	current_mode = MODE_SIOSERVER;
	sioserver_command_line = UART_MSR_CTS;
	sioserver_command_line_delta = UART_MSR_DCTS;

	set_lcr(standard_lcr, 0);
	set_baudrate(default_baudrate, 0);
	serial_config.IER |= UART_IER_MSI;
	serial_out(UART_IER, serial_config.IER);

	if (do_locks) {
		spin_unlock(&serial_config_lock);
		spin_unlock_irqrestore(&uart_lock, flags);
	}

	reset_cmdframe_buf(1);
}


static int get_command_frame(unsigned long arg)
{
	wait_queue_t wait;
	SIO_command_frame frame;
	signed long expire = HZ;
	signed long end_time = jiffies + HZ;
	unsigned long flags;

again:
	expire = end_time - jiffies;
	if (expire <= 0) {
		return -EATARISIO_COMMAND_TIMEOUT;
	}

	init_waitqueue_entry(&wait, current);

	add_wait_queue(&cmdframe_queue, &wait);
	while (1) {
		current->state = TASK_INTERRUPTIBLE;
		if ( cmdframe_buf.is_valid ) {
			PRINT_TIMESTAMP("found valid command frame");
			DBG_PRINTK(DEBUG_NOISY, "found valid command frame\n");
			break;
		}
		expire = schedule_timeout(expire);
		if (expire==0) {
			break;
		}
		if (signal_pending(current)) {
			current->state=TASK_RUNNING;
			remove_wait_queue(&cmdframe_queue, &wait);
			DBG_PRINTK(DEBUG_STANDARD, "got signal while waiting for command frame\n");
			return -EINTR;
		}
	}
	current->state=TASK_RUNNING;
	remove_wait_queue(&cmdframe_queue, &wait);

	if (!cmdframe_buf.is_valid ) {
		DBG_PRINTK(DEBUG_STANDARD, "waiting for command frame timed out\n");
		return -EATARISIO_COMMAND_TIMEOUT;
	}
	spin_lock_irqsave(&cmdframe_lock, flags);
/*
	if (!cmdframe_buf.is_valid) {
		DBG_PRINTK(DEBUG_STANDARD, "command frame invalidated shortly after notification\n");
		spin_unlock_irqrestore(&cmdframe_lock, flags);
		goto again;
	}
*/

	if (check_command_frame_time(0)) {
		cmdframe_buf.is_valid = 0;
		spin_unlock_irqrestore(&cmdframe_lock, flags);
		goto again;
	}

	frame.device_id   = cmdframe_buf.buf[0];
	frame.command     = cmdframe_buf.buf[1];
	frame.aux1        = cmdframe_buf.buf[2];
	frame.aux2        = cmdframe_buf.buf[3];
	frame.missed_count = cmdframe_buf.missed_count;
	cmdframe_buf.missed_count=0;
	current_cmdframe_serial_number = cmdframe_buf.serial_number;
	cmdframe_buf.is_valid = 0; /* already got that :-) */
	reset_tx_buf();
	reset_rx_buf();
	spin_unlock_irqrestore(&cmdframe_lock, flags);

	if (copy_to_user((SIO_command_frame*) arg, &frame, sizeof(SIO_command_frame)) ) {
		return -EFAULT;
	}
	return 0;
}

static int send_single_character(unsigned char c)
{
	int w;
	tx_buf.buf[tx_buf.head]=c;
	tx_buf.head = (tx_buf.head + 1) % IOBUF_LENGTH;
	initiate_send();
	w=wait_send(1);
	return w;
}


static int perform_send_data_frame(unsigned long arg)
{
	SIO_data_frame frame;
	int ret,w;

	if (copy_from_user(&frame, (SIO_data_frame*) arg, sizeof(SIO_data_frame)) ) {
		return -EFAULT;
	}

	udelay(DELAY_T3_MIN);


	if (frame.data_length) {
		if (add_highspeedpause && (serial_config.baudrate != ATARISIO_STANDARD_BAUDRATE)) {
			set_lcr(slow_lcr, 1);
		}
		if ((ret=setup_send_data_frame(frame.data_length, frame.data_buffer)) ) {
			if (add_highspeedpause && (serial_config.baudrate != ATARISIO_STANDARD_BAUDRATE)) {
				set_lcr(standard_lcr, 1);
			}
			return ret;
		}

		PRINT_TIMESTAMP("perform_send_data_frame: initiate_send");
		initiate_send();

		PRINT_TIMESTAMP("perform_send_data_frame: begin wait_send");
		if ((w=wait_send(frame.data_length+1))) {
			if (add_highspeedpause && (serial_config.baudrate != ATARISIO_STANDARD_BAUDRATE)) {
				set_lcr(standard_lcr, 1);
			}
			DBG_PRINTK(DEBUG_STANDARD, "send data frame returned %d\n",w);
			return w;
		}
		if (add_highspeedpause && (serial_config.baudrate != ATARISIO_STANDARD_BAUDRATE)) {
			set_lcr(standard_lcr, 1);
		}
		PRINT_TIMESTAMP("perform_send_data_frame: end wait_send");
	}
	return 0;
}

static int perform_receive_data_frame(unsigned long arg)
{
	SIO_data_frame frame;
	int ret=0;
	int received_len, copy_len;
	unsigned char checksum;

	if (copy_from_user(&frame, (SIO_data_frame*) arg, sizeof(SIO_data_frame)) ) {
		return -EFAULT;
	}


	if (frame.data_length) {
		/* receive data block and checksum*/
		received_len = wait_receive(frame.data_length+1, DELAY_T3_MAX/1000);

		if (received_len < 0) {
			DBG_PRINTK(DEBUG_STANDARD, "receive data block returned %d\n",received_len);
			return received_len;
		}
		copy_len = frame.data_length;
		if (received_len < copy_len) {
			copy_len = received_len;
		}
		if (copy_len > 0) {
			if ( (ret=copy_received_data_to_user(copy_len, frame.data_buffer)) ) {
				return ret;
			}
		}

	        if ((unsigned int) received_len < frame.data_length+1) {
			DBG_PRINTK(DEBUG_STANDARD, "receive data frame timed out [wanted %d got %d]\n",
					frame.data_length+1, received_len);
			return -EATARISIO_COMMAND_TIMEOUT;
		}
		checksum = calculate_checksum((unsigned char*)(rx_buf.buf),
			       	rx_buf.tail, frame.data_length, IOBUF_LENGTH);

		if (rx_buf.buf[(rx_buf.tail+frame.data_length) % IOBUF_LENGTH ]
			!= checksum) {
			DBG_PRINTK(DEBUG_STANDARD, "checksum error : 0x%02x != 0x%02x\n",
				checksum,
				rx_buf.buf[(rx_buf.tail+frame.data_length) % IOBUF_LENGTH]);
			if (! ret) {
				ret = -EATARISIO_CHECKSUM_ERROR;
			}
		}


		rx_buf.tail = (rx_buf.tail + frame.data_length + 1 ) % IOBUF_LENGTH;

		udelay(DELAY_T4_MIN);
		if (ret == 0) {
			ret=send_single_character(DATA_FRAME_ACK_CHAR);
		} else {
			send_single_character(DATA_FRAME_NAK_CHAR);
		}
	} else {
		return -EINVAL;
	}

	if (rx_buf.head != rx_buf.tail) {
		DBG_PRINTK(DEBUG_NOISY, "detected excess characters\n");
	}
	return ret;
}

static void print_status(void)
{
	unsigned long flags;

	switch (current_mode) {
	case MODE_1050_2_PC:
		PRINTK("current_mode: 1050-2-PC\n"); break;
	case MODE_SIOSERVER:
		PRINTK("current_mode: sioserver\n"); break;
	default:
		PRINTK("current_mode: unknown - %d\n", current_mode);
		break;
	}

	PRINTK("rx_buf: head=%d tail=%d wakeup_len=%d\n",
		rx_buf.head, rx_buf.tail, rx_buf.wakeup_len);

	PRINTK("tx_buf: head=%d tail=%d\n",
		tx_buf.head, tx_buf.tail);

	PRINTK("cmdframe_buf: valid=%d serial=%d missed=%d\n",
		cmdframe_buf.is_valid, 
		cmdframe_buf.serial_number,
		cmdframe_buf.missed_count);

	PRINTK("cmdframe: receiving=%d pos=%d error_counter=%d\n",
		cmdframe_buf.receiving,
		cmdframe_buf.pos, cmdframe_buf.error_counter);

	PRINTK("serial_config: baudrate=%d\n",
		serial_config. baudrate);

	PRINTK("command_line=0x%02x delta=0x%02x\n",
		sioserver_command_line,
		sioserver_command_line_delta);

	PRINTK("default_baudrate=%d do_autobaud=%d\n",
		default_baudrate, do_autobaud);

	spin_lock_irqsave(&uart_lock, flags);

	PRINTK("UART RBR = 0x%02x\n", serial_in(UART_RX));
	PRINTK("UART IER = 0x%02x\n", serial_in(UART_IER));
	PRINTK("UART IIR = 0x%02x\n", serial_in(UART_IIR));
	PRINTK("UART LCR = 0x%02x\n", serial_in(UART_LCR));
	PRINTK("UART MCR = 0x%02x\n", serial_in(UART_MCR));
	PRINTK("UART LSR = 0x%02x\n", serial_in(UART_LSR));
	PRINTK("UART MSR = 0x%02x\n", serial_in(UART_MSR));

	spin_unlock_irqrestore(&uart_lock, flags);
}

static int atarisio_ioctl(struct inode* inode, struct file* f,
 	unsigned int cmd, unsigned long arg)
{
	/*
	 * check for illegal commands
	 */
	int ret=0;

	if (_IOC_TYPE(cmd) != ATARISIO_IOC_MAGIC) return -ENOTTY;
	if (_IOC_NR(cmd) > ATARISIO_IOC_MAXNR) return -ENOTTY;

        DBG_PRINTK(DEBUG_VERY_NOISY, "ioctl 0x%x , 0x%lx\n", cmd, arg);

	switch (cmd) {
	case ATARISIO_IOC_GET_VERSION:
		ret = ATARISIO_VERSION;
		break;
	case ATARISIO_IOC_SET_MODE:
		switch (arg) {
		case ATARISIO_MODE_1050_2_PC:
			set_1050_2_pc_mode(1);
			break;
		case ATARISIO_MODE_PROSYSTEM:
			set_prosystem_mode(1);
			break;
		case ATARISIO_MODE_SIOSERVER:
			set_sioserver_mode(1);
			break;
		case ATARISIO_MODE_SIOSERVER_DSR:
			set_sioserver_mode_dsr(1);
			break;
		case ATARISIO_MODE_SIOSERVER_CTS:
			set_sioserver_mode_cts(1);
			break;
		default:
			return -EINVAL;
		}
		break;
	case ATARISIO_IOC_SET_BAUDRATE:
		default_baudrate = arg;
		ret = set_baudrate(default_baudrate, 1);
		break;
	case ATARISIO_IOC_SET_AUTOBAUD:
		do_autobaud = arg;
		if (do_autobaud == 0) {
			ret = set_baudrate(default_baudrate, 1);
		}
		break;
	case ATARISIO_IOC_DO_SIO:
		ret = perform_sio((SIO_parameters*)arg);
		if (ret == 0) {
			DBG_PRINTK(DEBUG_NOISY, "sio successful\n");
		} else {
			DBG_PRINTK(DEBUG_NOISY, "sio failed : %d\n", ret);
		}
		break;
	case ATARISIO_IOC_GET_COMMAND_FRAME:
		PRINT_TIMESTAMP("start getting command frame");
		ret = get_command_frame(arg);
		PRINT_TIMESTAMP("end getting command frame");
		break;
	case ATARISIO_IOC_SEND_COMMAND_ACK:
	case ATARISIO_IOC_SEND_COMMAND_ACK_XF551:
		if ((ret = check_new_command_frame())) {
			break;
		}
		if ((ret=check_command_frame_time(1))) {
			break;
		}
		PRINT_TIMESTAMP("start sending command ACK");
		ret = send_single_character(COMMAND_FRAME_ACK_CHAR);
		PRINT_TIMESTAMP("end sending command ACK");
		if (cmd == ATARISIO_IOC_SEND_COMMAND_ACK_XF551) {
			set_baudrate(ATARISIO_XF551_BAUDRATE, 1);
		}
		break;
	case ATARISIO_IOC_SEND_COMMAND_NAK:
		if ((ret = check_new_command_frame())) {
			break;
		}
		if ((ret=check_command_frame_time(1))) {
			break;
		}
		ret = send_single_character(COMMAND_FRAME_NAK_CHAR);
		break;
	case ATARISIO_IOC_SEND_DATA_ACK:
		udelay(DELAY_T4_MIN);
		ret = send_single_character(COMMAND_FRAME_ACK_CHAR);
		break;
	case ATARISIO_IOC_SEND_DATA_NAK:
		udelay(DELAY_T4_MIN);
		ret = send_single_character(COMMAND_FRAME_NAK_CHAR);
		break;
	case ATARISIO_IOC_SEND_COMPLETE:
	case ATARISIO_IOC_SEND_COMPLETE_XF551:
		if ((ret = check_new_command_frame())) {
			break;
		}
		PRINT_TIMESTAMP("start sending complete");
		if (add_highspeedpause) {
			udelay(DELAY_T5_MIN_SLOW);
		} else {
			udelay(DELAY_T5_MIN);
		}
		ret = send_single_character(OPERATION_COMPLETE_CHAR);
		PRINT_TIMESTAMP("end sending complete");
		if (cmd == ATARISIO_IOC_SEND_COMPLETE_XF551) {
			set_baudrate(ATARISIO_STANDARD_BAUDRATE, 1);
		}
		break;
	case ATARISIO_IOC_SEND_ERROR:
		if ((ret = check_new_command_frame())) {
			break;
		}
		if (add_highspeedpause) {
			udelay(DELAY_T5_MIN_SLOW);
		} else {
			udelay(DELAY_T5_MIN);
		}
		ret = send_single_character(OPERATION_ERROR_CHAR);
		break;
	case ATARISIO_IOC_SEND_DATA_FRAME:
	case ATARISIO_IOC_SEND_DATA_FRAME_XF551:
		if ((ret = check_new_command_frame())) {
			break;
		}
		PRINT_TIMESTAMP("begin send data frame");
		ret = perform_send_data_frame(arg);
		PRINT_TIMESTAMP("end send data frame");
		if (cmd == ATARISIO_IOC_SEND_DATA_FRAME_XF551) {
			set_baudrate(ATARISIO_STANDARD_BAUDRATE, 1);
		}
		break;
	case ATARISIO_IOC_RECEIVE_DATA_FRAME:
		if ((ret = check_new_command_frame())) {
			break;
		}
		ret = perform_receive_data_frame(arg);
		break;
	case ATARISIO_IOC_GET_BAUDRATE:
		ret = serial_config.baudrate;
		break;
	case ATARISIO_IOC_SET_HIGHSPEEDPAUSE:
		add_highspeedpause = arg;
		break;
	case ATARISIO_IOC_PRINT_STATUS:
		print_status();
		ret = 0;
		break;
	default:
		return -EINVAL;
	}
	return ret;
}

static unsigned int atarisio_poll(struct file* filep, poll_table* wait)
{
	unsigned int mask=0;

#ifdef ATARISIO_EARLY_NOTIFICATION
	unsigned long flags;
	unsigned long long current_time;
#endif

	poll_wait(filep, &cmdframe_queue, wait);

#ifdef ATARISIO_EARLY_NOTIFICATION
	poll_wait(filep, &early_cmdframe_queue, wait);

	spin_lock_irqsave(&cmdframe_lock, flags);

	current_time = get_timestamp();
	if (cmdframe_buf.receiving) {
		if (current_time > cmdframe_buf.start_reception_time + 10000) {
			cmdframe_buf.receiving = 0;
			DBG_PRINTK(DEBUG_STANDARD, "aborted reception of command frame - maximum time exceeded");
		} else {
			mask |= POLLPRI;
		}
	}
	spin_unlock_irqrestore(&cmdframe_lock, flags);
#endif

	if (cmdframe_buf.is_valid) {
		mask |= POLLPRI;
	}

	return mask;
}

static int busy=0;

static int atarisio_open(struct inode* inode, struct file* f)
{
	unsigned long flags;

	if (busy) return -EBUSY;
	busy=1;

	current_mode = MODE_1050_2_PC;

	reset_rx_buf();
	reset_tx_buf();
	reset_cmdframe_buf(1);

	do_autobaud = 0;
	default_baudrate = ATARISIO_STANDARD_BAUDRATE;
	add_highspeedpause = 0;

	spin_lock_irqsave(&uart_lock, flags);
	spin_lock(&serial_config_lock);

	serial_config.baud_base = 115200;
	serial_config.baudrate = default_baudrate;
	serial_config.IER = UART_IER_RDI;
	serial_config.MCR = UART_MCR_OUT2;
			/* OUT2 is essential for enabling interrupts - when it's required :-) */
	serial_config.LCR = standard_lcr; /* 1 stopbit, no parity */

	serial_config.FCR = UART_FCR_ENABLE_FIFO | UART_FCR_TRIGGER_1;

	serial_out(UART_IER, 0);

	serial_out(UART_LCR, serial_config.LCR);

	set_baudrate(serial_config.baudrate, 0);

	set_1050_2_pc_mode(0); /* this call sets MCR */

	serial_out(UART_FCR, 0);
	serial_out(UART_FCR, UART_FCR_ENABLE_FIFO | UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT);
	serial_out(UART_FCR, UART_FCR_ENABLE_FIFO | UART_FCR_TRIGGER_1);


	/*
	 * clear any interrupts
	 */
        (void)serial_in(UART_LSR);
	(void)serial_in(UART_IIR);
	(void)serial_in(UART_RX);
	(void)serial_in(UART_MSR);


	if (request_irq(irq, atarisio_interrupt, SA_SHIRQ | SA_INTERRUPT, NAME, &io)) {
		if (request_irq(irq, atarisio_interrupt, SA_SHIRQ, NAME, &io)) {
			PRINTK("could not register interrupt %d\n",irq);
			busy=0;
			spin_unlock_irqrestore(&uart_lock, flags);
			return -EFAULT;
		} else {
			DBG_PRINTK(DEBUG_STANDARD, "got slow interrupt\n");
		}
	} else {
		DBG_PRINTK(DEBUG_STANDARD, "got fast interrupt\n");
	}

	serial_out(UART_IER, serial_config.IER);

	spin_unlock(&serial_config_lock);
	spin_unlock_irqrestore(&uart_lock, flags);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
        MOD_INC_USE_COUNT;
#endif
	return 0;
}

static int atarisio_release(struct inode* inode, struct file* f)
{
	unsigned long flags;

	spin_lock_irqsave(&uart_lock, flags);

	serial_out(UART_MCR,0);
	serial_out(UART_IER,0);

	spin_unlock_irqrestore(&uart_lock, flags);

	free_irq(irq,&io);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
        MOD_DEC_USE_COUNT;
#endif

	busy=0;
	return 0;
}

static struct file_operations atarisio_fops = {
	poll:		atarisio_poll,
	ioctl:		atarisio_ioctl,
	open:		atarisio_open,
	release:	atarisio_release
};

static struct miscdevice atarisio_dev = {
	ATARISIO_DEFAULT_MINOR,
	NAME,
	&atarisio_fops
};

static struct serial_struct orig_serial_struct;
static int need_reenable = 0;

static int disable_serial_port(void)
{
	mm_segment_t fs;
	struct file* f;
	struct serial_struct ss;

	lock_kernel();

	fs = get_fs();
	set_fs(get_ds());

	f = filp_open(saved_port,O_RDWR,0);
	if (IS_ERR(f)) {
		DBG_PRINTK(DEBUG_STANDARD, "error opening serial port %s\n", saved_port);
		goto fail;
	}
	if (f->f_op && f->f_op->ioctl) {
		if (f->f_dentry && f->f_dentry->d_inode) {
			if (f->f_op->ioctl(f->f_dentry->d_inode, f, TIOCGSERIAL, (unsigned long) &ss)) {
				DBG_PRINTK(DEBUG_STANDARD, "TIOCGSERIAL failed\n");
				goto fail_close;
			}
			if (f->f_op->ioctl(f->f_dentry->d_inode, f, TIOCGSERIAL, (unsigned long) &orig_serial_struct)) {
				DBG_PRINTK(DEBUG_STANDARD, "TIOCGSERIAL failed\n");
				goto fail_close;
			}
			switch (ss.type) {
			case PORT_16550:
			case PORT_16550A:
				break;
			default:
				PRINTK("illegal port type - only 16550(A) is supported\n");
				goto fail_close;
			}

			DBG_PRINTK(DEBUG_STANDARD, "ss.port = 0x%04x ss.irq = %d ss.type = %d\n",
				ss.port, ss.irq, ss.type);

			if (io == 0) {
				io = ss.port;
			}
			if (irq == 0) {
				irq = ss.irq;
			}

			// disable serial driver by setting the uart type to none
			ss.type = PORT_UNKNOWN;

			if (f->f_op->ioctl(f->f_dentry->d_inode, f, TIOCSSERIAL, (unsigned long) &ss)) {
				DBG_PRINTK(DEBUG_STANDARD, "TIOCSSERIAL failed\n");
				goto fail_close;
			}
			need_reenable = 1;
		} else {
			DBG_PRINTK(DEBUG_STANDARD, "unable to get inode of %s\n", saved_port);
			goto fail_close;
		}
	} else {
		DBG_PRINTK(DEBUG_STANDARD, "device doesn't provide ioctl function!\n");
		goto fail_close;
	}

	if (filp_close(f,NULL)) {
		DBG_PRINTK(DEBUG_STANDARD, "error closing serial port %s\n", saved_port);
		goto fail;
	}
	set_fs(fs);
	unlock_kernel();
	return 0;

fail_close:
	if (filp_close(f,NULL)) {
		DBG_PRINTK(DEBUG_STANDARD, "error closing serial port %s\n", saved_port);
		goto fail;
	}

fail:
	set_fs(fs);
	unlock_kernel();
	return 1;
}

static int reenable_serial_port(void)
{
	mm_segment_t fs;
	struct file* f;

	if (!need_reenable) {
		DBG_PRINTK(DEBUG_STANDARD, "unnecessary call to reenable_serial_port\n");
		return 0;
	}

	lock_kernel();
	fs = get_fs();
	set_fs(get_ds());

	f = filp_open(saved_port,O_RDWR,0);
	if (IS_ERR(f)) {
		DBG_PRINTK(DEBUG_STANDARD, "error opening serial port %s\n", saved_port);
		goto fail;
	}
	if (f->f_op && f->f_op->ioctl) {
		if (f->f_dentry && f->f_dentry->d_inode) {
			if (f->f_op->ioctl(f->f_dentry->d_inode, f, TIOCSSERIAL, (unsigned long) &orig_serial_struct)) {
				DBG_PRINTK(DEBUG_STANDARD, "TIOCSSERIAL failed\n");
				goto fail_close;
			}
		} else {
			DBG_PRINTK(DEBUG_STANDARD, "unable to get inode of %s\n", saved_port);
			goto fail_close;
		}
	} else {
		DBG_PRINTK(DEBUG_STANDARD, "device doesn't provide ioctl function!\n");
		goto fail_close;
	}

	if (filp_close(f,NULL)) {
		DBG_PRINTK(DEBUG_STANDARD, "error closing serial port %s\n", saved_port);
		goto fail;
	}
	set_fs(fs);
	unlock_kernel();
	return 0;

fail_close:
	if (filp_close(f,NULL)) {
		DBG_PRINTK(DEBUG_STANDARD, "error closing serial port %s\n", saved_port);
		goto fail;
	}

fail:
	set_fs(fs);
	unlock_kernel();
	return 1;
}

static int atarisio_is_initialized = 0;

static int atarisio_init_module(void)
{
	int ret = 0;
	atarisio_is_initialized = 0;
	if (port != NULL) {
		saved_port = (char*) kmalloc(strlen(port)+1, GFP_KERNEL);
		if (saved_port == NULL) {
			return -ENOMEM;
		}
		strcpy(saved_port, port);
		if (disable_serial_port()) {
			PRINTK("unable to disable serial port %s\n", port);
			return -EINVAL;
		} else {
			DBG_PRINTK(DEBUG_STANDARD, "successfully disabled serial port %s\n", saved_port);
		}
	} else {
		saved_port = 0;
	}

	if ( ( io == 0) || (irq == 0) ) {
		PRINTK("please supply port or io and irq parameters\n");
		return -EINVAL;
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	if (check_region(io,8)) {
		PRINTK("cannot access IO-ports 0x%04x-0x%04x\n",io,io+7);
		ret = -EINVAL;
		goto failure;
	}
	request_region(io, 8, NAME);
#else
	if (request_region(io, 8, NAME) == 0) {
		PRINTK("cannot access IO-ports 0x%04x-0x%04x\n",io,io+7);
		ret = -EINVAL;
		goto failure;
	}
#endif

	
	/*
	 * check for 16550:
	 * this is just a very simple test using the scratch register
	 */
	serial_out(UART_SCR,0xaa);
	if (serial_in(UART_SCR) != 0xaa) {
		PRINTK("couldn't detect 16550\n");
		ret = -ENODEV;
		goto failure_release;
	}

	serial_out(UART_SCR,0x55);
	if (serial_in(UART_SCR) != 0x55) {
		PRINTK("couldn't detect 16550\n");
		ret = -ENODEV;
		goto failure_release;
	}

	/*
	 * checks are OK so far, try to register the device
	 */
	init_waitqueue_head(&rx_queue);
	init_waitqueue_head(&tx_queue);
	init_waitqueue_head(&cmdframe_queue);

#ifdef ATARISIO_EARLY_NOTIFICATION
	init_waitqueue_head(&early_cmdframe_queue);
#endif

	if (minor != ATARISIO_DEFAULT_MINOR) {
		atarisio_dev.minor = minor;
		PRINTK("using miscdevice minor number %d\n", minor);
	}
	if (misc_register(&atarisio_dev))
	{
		PRINTK("failed to register device\n");
		ret = -ENODEV;
		goto failure_release;
	}

	PRINTK("module successfully loaded io=0x%04x, irq=%d\n", io, irq);
	atarisio_is_initialized = 1;
	return 0;

failure_release:
	release_region(io, 8);

failure:
	if (need_reenable) {
		if (reenable_serial_port()) {
			PRINTK("error re-enabling serial port!\n");
		} else {
			DBG_PRINTK(DEBUG_STANDARD, "successfully re-enabled serial port %s\n", port);
		}
	}
	if (saved_port) {
		kfree(saved_port);
		saved_port = 0;
	}
	return ret;
}
	
static void atarisio_cleanup_module(void)
{
	if (!atarisio_is_initialized) {
		PRINTK("internal error - atarisio is not initialized!\n");
		return;
	}
	if (misc_deregister(&atarisio_dev)) {
		PRINTK("cannot unregister device\n");
	}

	release_region(io,8);
	if (need_reenable) {
		if (reenable_serial_port()) {
			PRINTK("error re-enabling serial port!\n");
		} else {
			DBG_PRINTK(DEBUG_STANDARD, "successfully re-enabled serial port %s\n", saved_port);
		}
	}
	if (saved_port) {
		kfree(saved_port);
		saved_port = 0;
	}
	PRINTK("module terminated\n");
	atarisio_is_initialized = 0;
}  

module_init(atarisio_init_module);
module_exit(atarisio_cleanup_module);

