/*
   dir2atr - create ATR image from directory

   Copyright (C) 2004 Matthias Reichl <hias@horus.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>

#include "AtrMemoryImage.h"
#include "SIOTracer.h"
#include "FileTracer.h"
#include "Dos2xUtils.h"
#include "MiscUtils.h"
#include "VirtualImageObserver.h"

#ifdef ALL_IN_ONE
int dir2atr_main(int argc, char**argv)
#else
int main(int argc, char**argv)
#endif
{
	RCPtr<AtrMemoryImage> image;
	RCPtr<Dos2xUtils> dos2xutils;
	RCPtr<VirtualImageObserver> observer;

	SIOTracer* sioTracer = SIOTracer::GetInstance();
	{
		RCPtr<FileTracer> tracer(new FileTracer(stderr));
		sioTracer->AddTracer(tracer);
		sioTracer->SetTraceGroup(SIOTracer::eTraceInfo, true, tracer);
		sioTracer->SetTraceGroup(SIOTracer::eTraceDebug, true, tracer);
	}


	bool dd = false;
	bool mydos = false;
	bool piconame = false;
	bool autoSectors = false;
	int sectors;
	char* directory;
	char* atrfilename;
	ESectorLength seclen;

	Dos2xUtils::EBootSectorType bootsecType = Dos2xUtils::eBootsecDefault;

	char c;
	while ( (c = getopt(argc, argv, "dmpb:")) != -1) {
		switch(c) {
		case 'd': dd = true; printf("using dd\n"); break;
		case 'm': mydos = true; printf("using mydos format\n"); break;
		case 'p': piconame = true; printf("creating PICONAME.TXT\n"); break;
		case 'b':
			if (strcasecmp(optarg,"dos20") == 0) {
				bootsecType = Dos2xUtils::eBootsecDos20;
			} else if (strcasecmp(optarg,"dos25") == 0) {
				bootsecType = Dos2xUtils::eBootsecDos25;
			} else if (strcasecmp(optarg,"mydos453") == 0) {
				bootsecType = Dos2xUtils::eBootsecMyDos453;
			} else if (strcasecmp(optarg,"turbodos21") == 0) {
				bootsecType = Dos2xUtils::eBootsecTurboDos21;
			} else if (strcasecmp(optarg,"turbodos21hs") == 0) {
				bootsecType = Dos2xUtils::eBootsecTurboDos21HS;
			} else if (strcasecmp(optarg,"mypicodos403") == 0) {
				bootsecType = Dos2xUtils::eBootsecMyPicoDos403;
			} else if (strcasecmp(optarg,"mypicodos403hs") == 0) {
				bootsecType = Dos2xUtils::eBootsecMyPicoDos403HS;
			} else {
				printf("unknown boot sector type \"%s\"\n", optarg);
			}
		}
	}

	if (dd) {
		seclen = e256BytesPerSector;
	} else {
		seclen = e128BytesPerSector;
	}

	if (argc == optind+2) {
		autoSectors = true;
	} else if (argc != optind+3) {
		goto usage;
	}

	if (!autoSectors) {
		sectors = atoi(argv[optind++]);
		if (sectors < 720 || sectors > 65535) {
			printf("illegal sector number - must be 720..65535\n");
			return 1;
		}
	}

	atrfilename=argv[optind++];
	directory=argv[optind++];

	if (autoSectors) {
		if (!mydos) {
			printf("number of sectors not specified - using MyDOS format\n");
			mydos = true;
		}

		sectors = MiscUtils::EstimateDiskSize(directory, seclen, piconame);
		printf("calculated disk size is %d sectors\n", sectors);
	}

	image = new AtrMemoryImage;
	if (dd) {
		if (!image->CreateImage(seclen, sectors)) {
			printf("cannot create atr image\n");
			return 1;
		}
	} else {
		if (!image->CreateImage(seclen, sectors)) {
			printf("cannot create atr image\n");
			return 1;
		}
	}

	observer = new VirtualImageObserver(image);
	dos2xutils = new Dos2xUtils(image, directory, observer.GetRealPointer());
	observer->SetRootDirectoryObserver(dos2xutils);

	if (mydos) {
		if (!dos2xutils->SetDosFormat(Dos2xUtils::eMyDos)) {
			printf("cannot set MyDos format\n");
			return 1;
		}
	} else {
		if (!dos2xutils->SetDosFormat(Dos2xUtils::eDos2x)) {
			printf("cannot set DOS 2.x format\n");
			return 1;
		}
	}
	if (!dos2xutils->InitVTOC()) {
		printf("creating directory failed\n");
		return 1;
	}

	dos2xutils->AddFiles(piconame);

	if (!dos2xutils->WriteBootSectors(bootsecType)) {
		printf("writing boot sectors failed\n");
	}

	if (!image->WriteImageToFile(atrfilename)) {
		printf("error writing image to \"%s\"\n", atrfilename);
	}

	dos2xutils = 0;
	observer = 0;

	SIOTracer::GetInstance()->RemoveAllTracers();
	return 0;
usage:
	printf("dir2atr V0.01 (c) 2004-2005 by Matthias Reichl\n");
	printf("usage: dir2atr [-d] [-m] [-p] [-b <DOS>] [sectors] atrfile directory\n");
	printf("  -d        create double density image (default: single density)\n");
	printf("  -m        create MyDOS image (default: DOS 2.x)\n");
	printf("  -p        create PICONAME.TXT (long filename description)\n");
	printf("  -b <DOS>  write boot sectors for specified DOS\n");
	printf("            Supported DOS are: Dos20, Dos25, MyDos453,\n");
	printf("            TurboDos21, TurboDos21HS, MyPicoDos403, MyPicoDos403HS\n");

	SIOTracer::GetInstance()->RemoveAllTracers();
	return 1;
}
